var EntityManager = require('../../lib/ecs/EntityManager');

QUnit.module('Entity and EntityManager');

test('entities[] array for adding and removing', function() {

    var m = new EntityManager();

    var e = m.addNewEntity();
    equal(m.entities.length, 1, 'added to array');
    equal(m.hasEntity(e), true, 'hasEntity');
    strictEqual(e._manager, m, 'manager set');

    m.removeEntity(e);
    equal(m.entities.length, 0, 'removed from array');
    equal(m.hasEntity(e), false, 'hasEntity');
    strictEqual(e._manager, null, 'manager cleared');

    // Same thing but via entity
    e = m.addNewEntity();
    e.remove();
    equal(m.entities.length, 0, 'removed from array');
    equal(m.hasEntity(e), false, 'hasEntity');
    strictEqual(e._manager, null, 'manager cleared');

});

test('tags handled on removal and adding', function() {

    var m = new EntityManager();
    var e = m.addNewEntity();

    m.tagEntity(e, 'tag');
    deepEqual(m.getEntitiesByTag('tag'), [e], 'getting tagged entity');
    deepEqual(e._tags, ['tag'], 'entity tag added');

    m.removeEntityTag(e, 'tag');
    deepEqual(m.getEntitiesByTag('tag'), [], 'removing entity tag');
    deepEqual(e._tags, [], 'entity tag removed');

    e = m.addNewEntity();
    m.tagEntity(e, 'foo');
    deepEqual(m.getEntitiesByTag('foo'), [e], 'getting tagged entity');
    m.removeEntity(e);
    deepEqual(m.getEntitiesByTag('foo'), [], 'removing entity tag');
    deepEqual(e._tags, [], 'entity tag removed');

    // Same thing but via entity
    e = m.addNewEntity();
    e.addTag('bar');
    deepEqual(m.getEntitiesByTag('bar'), [e], 'getting tagged entity');
    e.removeTag('bar');
    deepEqual(m.getEntitiesByTag('bar'), [], 'removing entity tag');

});

test('adding and removing components basics', function() {

    var m = new EntityManager();
    var e = m.addNewEntity();

    var C1 = function C1() { this.foo = 123; };

    m.addEntityComponent(e, C1);
    deepEqual(e._Components, [C1], 'component added to entity');
    strictEqual(e.C1.constructor, C1, 'comonent instance added');

    m.removeEntityComponent(e, C1);
    deepEqual(e._Components, [], 'component removed to entity');
    strictEqual(e.C1, undefined, 'comonent instance added');

    // Same thing but via entity
    e.addComponent(C1);
    deepEqual(e._Components, [C1], 'component added to entity');
    strictEqual(e.C1.constructor, C1, 'comonent instance added');
    e.removeComponent(C1);
    deepEqual(e._Components, [], 'component removed to entity');
    strictEqual(e.C1, undefined, 'comonent instance added');

});

test('creating indexes on on components', function() {

    var m  = new EntityManager();
    var e  = m.addNewEntity();
    function C1() { this.foo = 123; }
    function C2() { this.bar = 456; }

    e.addComponent(C1);
    m._createNewIndex([C1]);
    deepEqual(m.groupToEntitiesInfoMap.C1.entities, [e], 'entity in info');
    deepEqual(m.groupToEntitiesInfoMap.C1.Components, [C1], 'components in info');

    e.addComponent(C2);
    m._createNewIndex([C1,C2]);
    deepEqual(m.groupToEntitiesInfoMap.C1C2.entities, [e], 'entity in info');
    deepEqual(m.groupToEntitiesInfoMap.C1C2.Components, [C1, C2], 'components in info');

    var f = m.addNewEntity();
    f.addComponent(C1);
    deepEqual(m.getEntitiesByComponents([C1]), [e, f], 'entity added to existing index');
    f.addComponent(C2);
    deepEqual(m.getEntitiesByComponents([C1, C2]), [e, f], 'entity added to existing index');

    f.removeComponent(C1);
    deepEqual(m.getEntitiesByComponents([C1]), [e], 'entity added to existing index');
    deepEqual(m.getEntitiesByComponents([C1, C2]), [e], 'entity added to existing index');
    e.remove();
    deepEqual(m.getEntitiesByComponents([C1]), [], 'entity added to existing index');
    deepEqual(m.getEntitiesByComponents([C1, C2]), [], 'entity added to existing index');

});

test('Missing indexes', function() {

    var m  = new EntityManager();
    var e  = m.addNewEntity();
    function C1() { this.foo = 123; }
    function C2() { this.bar = 456; }

    e.addComponent(C1);
    e.addComponent(C2);
    m.getEntitiesByComponents([C1, C2]);
    deepEqual(m.groupToEntitiesInfoMap.C1C2.entities, [e], 'ad hoc index added');
    var f = m.addNewEntity();
    f.addComponent(C1);
    f.addComponent(C2);
    deepEqual(m.groupToEntitiesInfoMap.C1C2.entities, [e, f], 'ad hoc index added');
    e.remove(); f.remove();
    deepEqual(m.groupToEntitiesInfoMap.C1C2.entities, [], 'ad hoc index added');
});

test('Pool implementation', function() {

    var m  = new EntityManager();
    var e  = m.addNewEntity();
    e.remove();
    var e2 = m.addNewEntity();
    strictEqual(e, e2, 'object recycled');
    e2.remove();
    strictEqual(m.entityPool.freeList.length, 1, 'not increasing size');

});
